from typing import *
from collections import defaultdict
import argparse
import os
import pandas as pd
from tqdm import tqdm
import torch
from torch.utils.data import DataLoader
from ..scorer import IncrementalLMScorer, MaskedLMScorer

def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Utility to quickly retrieve per-token score for sentences using language models."
    )

    input_group = parser.add_mutually_exclusive_group(required=True)

    parser.add_argument(
        "--scorer",
        type=str,
        required=True,
        choices=['incremental', 'masked'],
        help="Language model type: either incremental or masked."
    )

    parser.add_argument(
        "--model",
        "-m",
        type=str,
        required=True,
        help="Name of the model (can use any model from the huggingface model hub https://huggingface.co/models)."
    )

    input_group.add_argument(
        "--text",
        type=str,
        help="Single sentence to score. Use --file instead for scoring multiple sentences."
    )

    input_group.add_argument(
        "--file",
        type=str,
        help="File containing one sentence per line."
    )

    parser.add_argument(
        "--batch-size",
        "-bs",
        type=int,
        default=1,
        help="Number of sentences to process in parallel."
    )

    parser.add_argument(
        "--num-workers",
        "-nw",
        type=int,
        default=2,
        help="Number of workers to spawn for scoring the sentences in the input file."
    )

    parser.add_argument(
        "--prob",
        action="store_true",
        help="Get per-token probabilities for every sentence in the input."
    )

    # parser.add_argument(
    #     "--logprob",
    #     "log-probability",
    #     default=True,
    #     action="store_true",
    #     help="Get per-token log probabilities for every sentence in the input."
    # )

    parser.add_argument(
        "--surprisal",
        action="store_true",
        help="Get per-token surprisals for every sentence in the input."
    )

    parser.add_argument(
        "--rank",
        action="store_true",
        help="Get per-token ranks (scored by log-probability) for every sentence in the input."
    )

    parser.add_argument(
        "--base-two",
        "--in_bits",
        action="store_true"
    )

    parser.add_argument(
        "--device",
        type=str,
        default="cpu",
        help="If provided, it runs the model on the specified cuda device."
    )

    return parser.parse_args()

def normalize_args(args: argparse.Namespace) -> None:
    if (args.file):
        args.file = os.path.realpath(args.file)

def validate_args(args: argparse.Namespace) -> None:
    if (args.file):
        if (not os.path.isfile(args.file)):
            raise ValueError("The provided sentences file path is invalid.")
        
        if (args.batch_size == 1 or args.num_workers == 2):
            print('Currently using batch size of {} and {} workers. Please configure them according to your hardware.'.format(args.batch_size, args.num_workers))
    
    if (args.device != 'cpu'):
        if (not torch.cuda.is_available()):
            raise ValueError("Cuda not available.")
        if (int(args.device.split(':')[-1]) >= torch.cuda.device_count()):
            raise ValueError("Invalid device.")

def pretty_print(results: Iterable) -> None:
    print(results.round(2).to_string(index=False))

def cli_token_score(self, batch, model_type, surprisal = False, prob = False, base_two = False, rank = False, batch_idx = 0, global_batch_size = 1):
    
    result = defaultdict(list)

    tokenized = self.prepare_text(batch)
    if rank:
        scores, ranks = self.compute_stats(tokenized, rank = rank, base_two = base_two, return_tensors=True)
    else:
        scores = self.compute_stats(tokenized, base_two = base_two, return_tensors=True)

    scores = [s.tolist() for s in scores]

    if model_type == 'incremental':
        indices = [[i for i in indexed if i != self.tokenizer.pad_token_id] for indexed in tokenized[0]['input_ids'].tolist()]
    else:
        indices = [[i.item() for i in indexed if i.item() != self.tokenizer.pad_token_id] for indexed in list(zip(*tokenized))[2]]

    tokens = [self.decode(idx) for idx in indices]

    if rank:
        for i, (t, s, r) in enumerate(zip(tokens, scores, ranks)):
            sent_id = i + batch_idx * global_batch_size + 1
            diff = 0
            if len(t) > len(s):
                diff = len(t) - len(s)
                s = [None]*diff + s
                r = [None]*diff + r
            result['sentence_id'].extend([sent_id]*len(t))
            result['token_id'].extend([j+1 for j, _ in enumerate(t)])
            result['token'].extend(t)
            result['logprob'].extend(s)
            
            if prob:
                probs = [None] * diff + torch.tensor(s[diff:]).exp().tolist()
                result['prob'].extend(probs)
            
            if surprisal:
                surps = [-1.0 * x if x is not None else x for x in s]
                result['surprisal'].extend(surps)

            result['rank'].extend(r)
    else:
        for i, (t, s) in enumerate(zip(tokens, scores)):
            sent_id = i + batch_idx * global_batch_size + 1
            diff = 0
            if len(t) > len(s):
                diff = len(t) - len(s)
                s = [None]*diff + s
            result['sentence_id'].extend([sent_id]*len(t))
            result['token_id'].extend([j+1 for j, _ in enumerate(t)])
            result['token'].extend(t)
            result['logprob'].extend(s)
            
            if prob:
                probs = [None] * diff + torch.tensor(s[diff:]).exp().tolist()
                result['prob'].extend(probs)
            
            if surprisal:
                surps = [-1.0 * x if x is not None else x for x in s]
                result['surprisal'].extend(surps)
    return result

def main(args: argparse.Namespace):
    modeltype = args.scorer.lower()
    if (modeltype == 'incremental'):
        lm = IncrementalLMScorer(args.model, args.device)
    elif (modeltype == 'masked'):
        lm = MaskedLMScorer(args.model, args.device)
    else:
        raise Exception("Incorrect scorer passed. Use either incremental or masked.")
    
    if (args.text):
        processed_results = cli_token_score(lm, args.text, modeltype, args.surprisal, args.prob, args.base_two, args.rank)
    else:
        sentence_file = open(args.file, 'r')
        sentences = [sentence.strip() for sentence in sentence_file]
        sentence_dl = DataLoader(sentences, batch_size=args.batch_size, num_workers=args.num_workers)
        results = []
        for i, batch in enumerate(tqdm(sentence_dl)):
            results.append(cli_token_score(lm, batch, modeltype, args.surprisal, args.prob, args.base_two, args.rank, batch_idx = i, global_batch_size=args.batch_size))
        
        processed_results = defaultdict(list)
        for result in results:
            for k, v in result.items():
                processed_results[k].extend(v)

    processed_results = pd.DataFrame(processed_results)
    pretty_print(processed_results)

def process():
    try:
        args = parse_args()
        validate_args(args)
        main(args)
    except KeyboardInterrupt:
        print('\nProcess stopped by user.')
    except Exception as er:
        print('Error: {}'.format(er))

if __name__ == "__main__":
    process()


from typing import Iterable, Union, List, Dict, Optional, Tuple

from transformers import AutoModelForSequenceClassification, AutoTokenizer
import torch

class SupervisedHead:
    """
    (Under construction)
    Implements the supervised head class to facilitate behavioral
    analyses of model outputs.
    """
    def __init__(self, model_name: str, device: Optional[str] = 'cpu') -> None:
        """
        :param model_name: name of the model, should either be a path
            to a model (.pt or .bin file) stored locally, or a
            pretrained model stored on the Huggingface Model Hub.
        :type model_name: str
        :param device: device type that the model should be loaded on,
            options: `cpu or cuda:{0, 1, ...}`
        :type device: str, optional
        """
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast = True)
        self.device = device
        self.model = AutoModelForSequenceClassification.from_pretrained(model_name, return_dict = True,)
        self.model.to(self.device)
        self.model.eval()

    def encode(self, inputs: Union[str, List[str]], return_tensors: Optional[str] = 'pt') -> Dict:
        """
        Encodes batch of inputs to the supervised model to return
        an encoded format to be passed to the model.

        :param ``Union[str, List[str]]`` inputs: batch of inputs to be
            encoded by the model.
        :param ``Optional[str]`` return_tensors: returned tensor format.
            Default `'pt'`
        :return: Dictionary of encoded input.
        """
        inputs = [inputs] if isinstance(inputs, str) else inputs

        encoded = self.tokenizer(inputs, padding = 'longest', return_tensors = 'pt')
        if self.device != 'cpu':
            encoded = encoded.to(self.device)
        
        return encoded
    
    def logits(self, inputs: Union[str, List[str]], probs: bool = True, verbose: bool = False) -> Union[torch.Tensor, Dict]:
        """
        Runs inference on the model and returns logits for each label
        depending on the supervised task on which the model was trained
        on.

        :param ``Union[str, List[str]]`` inputs: batch of inputs to be
            encoded by the model.
        :param probs: specifies whether to return probabilities.
        :type probs: bool
        :param verbose: specifies if the label names should be revealed
            in the output (if they exist).
        :type  verbose: bool
        :return: Either a torch tensor consisting of the model outputs
            or a dictionary consisting of `{label: probability}`.
        """
        encoded = self.encode(inputs)

        output = self.model(**encoded).logits.detach()

        if probs:
            output = output.softmax(1)

        if verbose:
            if 'label2id' not in self.model.config.to_dict().keys():
                output = output
            else:
                output = {k: output[:, v].tolist()for k, v in self.model.config.to_dict()['label2id'].items()}

        return output
        



from typing import Iterable, Union, List, Dict, Optional, Tuple

from .utils import find_pattern, find_index, find_paired_indices, character_span

from transformers import AutoModel, AutoTokenizer, AutoConfig
from transformers.utils.logging import set_verbosity_error
import torch

set_verbosity_error()

class CWE(object):
    """
    Implements the contextualized word embedding class to
    facilitate extraction of word representations form a given
    transformer model.

    :param model_name: name of the model, should either be a path
        to a model (.pt or .bin file) stored locally, or a
        pretrained model stored on the Huggingface Model Hub.
    :type model_name: str
    :param device: device type that the model should be loaded on,
        options: `cpu or cuda:{0, 1, ...}`
    :type device: str, optional
    :param pretrained: whether to load the model with pretrained weights.
        loads a randomly initialized model if `False`. Default = `True`.
    :type pretrained: bool
    """
    def __init__(self, model_name: str, device: Optional[str] = 'cpu', pretrained: bool = True) -> None:
        """
        :param model_name: name of the model, should either be a path
            to a model (.pt or .bin file) stored locally, or a
            pretrained model stored on the Huggingface Model Hub.
        :type model_name: str
        :param device: device type that the model should be loaded on,
            options: `cpu or cuda:{0, 1, ...}`
        :type device: str, optional
        :param pretrained: whether to load the model with pretrained weights.
            loads a randomly initialized model if `False`. Default = `True`.
        :type pretrained: bool
        """
        self.device = device
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast = True)
        self.pretrained = pretrained
        if pretrained:
            self.model = AutoModel.from_pretrained(model_name, return_dict = True, output_hidden_states = True)
        else:
            self.config = AutoConfig.from_pretrained(model_name, return_dict = True, output_hidden_states = True)
            self.model = AutoModel.from_config(self.config)
        self.dimensions = list(self.model.parameters())[-1].shape[0]

        self.layers = self.model.config.num_hidden_layers

        if self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"additional_special_tokens": ["<|pad|>"]})
            self.tokenizer.pad_token = "<|pad|>"
            self.model.resize_token_embeddings(len(self.tokenizer))
        
        self.model.to(self.device)
        self.model.eval()

    def encode_text(self, text: Union[str, List[str]], layer: Union[int, List[int]] = None) -> Tuple:
        """
        Encodes batch of raw sentences using the model to return hidden
        states at a given layer.

        :param ``Union[str, List[str]]`` text: batch of raw sentences
        :param layer: layer from which the hidden states are extracted.
        :type layer: int
        :return: Tuple `(input_ids, hidden_states)`
        """
        sentences = [text] if isinstance(text, str) else text

        # Encode sentence into ids stored in the model's embedding layer(s).
        encoded = self.tokenizer(sentences, padding = 'longest', return_tensors="pt")
        encoded = encoded.to(self.device)

        input_ids = encoded['input_ids']
        attention_mask = encoded['attention_mask'].unsqueeze(-1)
        if 'cuda' in self.device:
            attention_mask = attention_mask.cpu()

        if layer == 'static' or layer == 'pre':
            hidden_states = [self.model.embeddings.word_embeddings.weight.detach()[i] for i in input_ids]
            hidden_states = torch.stack(hidden_states)
            hidden_states = hidden_states * attention_mask
        else:
        # Compute hidden states for the sentence for the given layer.
            output = self.model(**encoded)

            # Hidden states appear as the last element of the otherwise custom hidden_states object
            if isinstance(layer, list) or layer == 'all':
                hidden_states = output.hidden_states
                if "cuda" in self.device:
                    input_ids = input_ids.cpu()
                    hidden_states = [h.detach().cpu() for h in hidden_states]
                else:
                    hidden_states = [h.detach() for h in hidden_states]
                if layer != 'all':
                    hidden_states = [hidden_states[i] for i in sorted(layer)]

                hidden_states = [h * attention_mask for h in hidden_states]
            else:
                # if layer != 'all':
                if layer is None:
                    layer = self.layers
                elif layer > self.layers:
                    raise ValueError(f"Number of layers specified ({layer}) exceed layers in model ({self.layers})!")
                hidden_states = output.hidden_states[layer]
                if "cuda" in self.device:
                    input_ids = input_ids.cpu()
                    hidden_states = hidden_states.detach().cpu()
                else:
                    hidden_states = hidden_states.detach()

                hidden_states = hidden_states * attention_mask
                # else:
                #     hidden_states = output.hidden_states
                
                #     if "cuda" in self.device:
                #         input_ids = input_ids.cpu()
                #         hidden_states = [h.detach().cpu() for h in hidden_states]
                #     else:
                #         hidden_states = [h.detach() for h in hidden_states]

        return input_ids, hidden_states
    # A function that extracts the representation of a given word in a sentence (first occurrence only)

    def extract_representation(self, sentence_words: Union[List[List[Union[str, Union[Tuple[int, int], str]]]], List[Union[str, Union[Tuple[int, int], str]]]], layer: Union[int, List[int]] = None) -> Union[torch.Tensor, List[torch.Tensor]]:
        """
        Extract representations from the model at a given layer.

        :param ``Union[List[List[Union[str, Union[Tuple(int, int), str]]]], List[Union[str, Union[Tuple(int, int), str]]]]`` sentence_words: Input 
            consisting of `[(sentence, word)]`, where sentence is an
            input sentence, and word is a word present in the sentence
            that will be masked out, or `[(sentence, (start, end))]`,
            where (start, end) is a tuple consisting of the character
            span indices that form the word.
        :param ``Union[int, List[int]]`` layer: layer(s) from which the hidden states are extracted.
        :return: torch tensors or list of torch
            tensors corresponding to word representations
        """
        sentences = [sentence_words] if isinstance(sentence_words[0], str) else sentence_words

        num_inputs = len(sentences)

        input_ids, hidden_states = self.encode_text(list(list(zip(*sentences))[0]), layer)

        if isinstance(sentences[0][1], str):
            sentences = [(s, character_span(s, w)) for s, w in sentences]
    
        search_queries = []
        for s, idx in sentences:
            if 0 in idx:
                search_queries.append(self.tokenizer.encode_plus(f'{s[idx[0]:idx[1]]}', add_special_tokens = False)['input_ids'])
            else:
                ## this one really matters if we are using GPT2
                search_queries.append(self.tokenizer.encode_plus(f' {s[idx[0]:idx[1]]}', add_special_tokens = False)['input_ids'])

        query_idx = list(map(lambda x: find_pattern(x[0], x[1]), zip(search_queries, input_ids.tolist())))

        if isinstance(layer, list) or layer == 'all':
            # representations = list(map(lambda x: x[torch.arange(num_inputs)[:, None], query_idx].mean(1), hidden_states))
            representations = list(map(lambda x: torch.stack([hs.squeeze()[idx[0]:idx[1]].mean(0) for hs, idx in zip(x.split([1] * num_inputs), query_idx)]), hidden_states))
        else:
            if layer is None:
                layer = self.layers
            elif layer == 'static' or 'pre':
                layer = layer
            elif layer > self.layers:
                raise ValueError(f"Number of layers specified ({layer}) exceed layers in model ({self.layers})!")
            # representations = hidden_states[torch.arange(num_inputs)[:, None], query_idx].mean(1)
            representations = torch.stack([hs.squeeze()[idx[0]:idx[1]].mean(0) for hs, idx in zip(hidden_states.split([1] * num_inputs), query_idx)])
        
        return representations

    def extract_sentence_representation(self, sentences: Union[str, List[str]], layer: Union[int, List[int]], pooler: str = "mean") -> Union[torch.Tensor, List[torch.Tensor]]:
        '''
        Extract representations of input sentences from one or more layers in the model.

        :param ``Union[str, List[str]]`` sentences: Input consisting of one or
            more sentences.
        :param ``Union[int, List[int]]`` layer: one or more layer from which the representations are
            extracted.
        :param pooler: pooling logic (mean, max, min)
        :type pooler: str

        :return: torch.Tensor
        '''
        
        sentences = [sentences] if isinstance(sentences, str) else sentences
        
        input_ids, hidden_states = self.encode_text(sentences, layer)

        lengths = torch.tensor([len([i for i in ids if i != 0]) for ids in input_ids.tolist()])

        if isinstance(hidden_states, list):
            hidden_states = torch.stack(hidden_states).mean(0)
        
        if pooler == "mean":
            hidden_states = torch.div(hidden_states.sum(1), lengths.view(-1, 1))
        elif pooler == 'min':
            hidden_states = hidden_states.min(0).values
        elif pooler == 'max':
            hidden_states = hidden_states.max(0).values
        else:
            print("Only pooler = mean, max, or min supported for now!")

        return hidden_states


    def extract_paired_representations(self, sentence_words: Union[Tuple[str], List[Tuple[str]]], layer:int = None) -> Tuple:
        raise NotImplementedError
        # '''
        # Extract representations of pairs of words from a given sentence
        # from the model at a given layer.

        # :param ``Union[Tuple[str], List[Tuple[str]]]`` text: Input 
        #     consisting of `[(sentence, word1, word2)]`, where sentence
        #     is an input sentence, and word1, word2 are two words
        #     present in the sentence that will be masked out.
        # :param layer: layer from which the representations are 
        #     extracted.
        # :type layer: int
        # :return: Tuple consisting of torch tensors or lists of torch
        #     tensors corresponding to word representations
        # '''
        # sentences = [sentence_words] if isinstance(sentence_words[0], str) else sentence_words

        # num_inputs = len(sentences)

        # input_ids, hidden_states = self.encode_text(list(list(zip(*sentences))[0]), layer)

        # if isinstance(sentences[0][1], str):
        #     sentences = [(s, *find_paired_indices(s, w1, w2)) for s, w1, w2 in sentences]
        
        # search_queries1 = []
        # for s, idx1, idx2 in sentences:
        #     if 0 in idx1:
        #         search_queries1.append(self.tokenizer.encode_plus(f'{" ".join(s.split()[idx1[0]:idx1[1]])}', add_special_tokens = False)['input_ids'])
        #     else:
        #         ## this one really matters if we are using GPT2
        #         search_queries1.append(self.tokenizer.encode_plus(f' {" ".join(s.split()[idx1[0]:idx1[1]])}', add_special_tokens = False)['input_ids'])

        # search_queries2 = []
        # for s, idx1, idx2 in sentences:
        #     if 0 in idx2:
        #         search_queries2.append(self.tokenizer.encode_plus(f'{" ".join(s.split()[idx2[0]:idx2[1]])}', add_special_tokens = False)['input_ids'])
        #     else:
        #         ## this one really matters if we are using GPT2
        #         search_queries2.append(self.tokenizer.encode_plus(f' {" ".join(s.split()[idx2[0]:idx2[1]])}', add_special_tokens = False)['input_ids'])

        # # search_queries1 = [self.tokenizer.encode_plus(f'{" ".join(s.split()[idx1[0]:idx1[1]])}', add_special_tokens = False)['input_ids'] for s, idx1, idx2 in sentences]
        # # search_queries2 = [self.tokenizer.encode_plus(f'{" ".join(s.split()[idx2[0]:idx2[1]])}', add_special_tokens = False)['input_ids'] for s, idx1, idx2 in sentences]

        # query_idx1 = list(map(lambda x: find_pattern(x[0], x[1]), zip(search_queries1, input_ids.tolist())))
        # query_idx2 = list(map(lambda x: find_pattern(x[0], x[1]), zip(search_queries2, input_ids.tolist())))

        # if layer != 'all':
        #     if layer is None:
        #         layer = self.layers
        #     elif layer > self.layers:
        #         raise ValueError(f"Number of layers specified ({layer}) exceed layers in model ({self.layers})!")
        #     representations1 = torch.stack([hs.squeeze()[idx[0]:idx[1]].mean(0) for hs, idx in zip(hidden_states.split([1] * num_inputs), query_idx1)])
        #     representations2 = torch.stack([hs.squeeze()[idx[0]:idx[1]].mean(0) for hs, idx in zip(hidden_states.split([1] * num_inputs), query_idx2)])
        # else:
        #     representations1 = list(map(lambda x: torch.stack([hs.squeeze()[idx[0]:idx[1]].mean(0) for hs, idx in zip(x.split([1] * num_inputs), query_idx1)]), hidden_states))
        #     representations2 = list(map(lambda x: torch.stack([hs.squeeze()[idx[0]:idx[1]].mean(0) for hs, idx in zip(x.split([1] * num_inputs), query_idx2)]), hidden_states))
        
        # return representations1, representations2


import random
import re
import string
import torch
from polyleven import levenshtein

from typing import List, Tuple, Optional

def get_batch(data: list, batch_size: int, shuffle: bool = False):
    if shuffle:
        random.shuffle(data)
    sindex = 0
    eindex = batch_size
    while eindex < len(data):
        batch = data[sindex:eindex]
        temp = eindex
        eindex = eindex + batch_size
        sindex = temp
        yield batch

    if eindex >= len(data):
        batch = data[sindex:]
        yield batch

def character_span(sentence, word):
    assert word in sentence
    idx = sentence.find(word)
    return idx, idx + len(word)

def find_pattern(pieces: List, whole: List) -> Tuple[int, int]:
    num_pieces = len(pieces)
    result = (0, 0)
    for i in (j for j,entry in enumerate(whole) if entry == pieces[0]):
        if whole[i:i+num_pieces] == pieces:
            result = (i, i + num_pieces)
    return result

def edit_distance(word1: str, word2: str) -> int:
    m, n = len(word1), len(word2) 
    dp = [[0 for x in range(n + 1)] for x in range(m + 1)] 

    for i in range(m + 1): 
        for j in range(n + 1): 
  
            if i == 0: 
                dp[i][j] = j    
            elif j == 0: 
                dp[i][j] = i    
            elif word1[i-1] == word2[j-1]: 
                dp[i][j] = dp[i-1][j-1] 
            else: 
                dp[i][j] = 1 + min(dp[i][j-1], dp[i-1][j], dp[i-1][j-1]) 
  
    return dp[m][n]

def argmin(lst: List) -> int:
    return min(range(len(lst)), key=lambda x: lst[x])

def argmax(lst: List) -> int:
    return max(range(len(lst)), key=lambda x: lst[x])
  
def find_index(context: str, word: str, method: Optional[str] = "regular") -> Tuple[int, int]:
    if method == "edit":
        tokenized = context.split()
        editdists = [edit_distance(w, word) for w in tokenized]
        
        index = argmin(editdists)
    else:
        # prefix, postfix = context.split(word)
        prefix, postfix = re.split(fr"\b{word}\b", context)
        word_length = len(word.split(" "))

        start = len(prefix.split())
        end = start + word_length

        #prefix = context.split(word)[0].strip().split()
        #index = len(prefix)
    
    return start, end

def gen_words(length: int) -> str:
    return " ".join([char for char in string.ascii_lowercase[0:length]])

def find_paired_indices(context: str, word1: str, word2: str, importance: int = 1) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    if importance == 1:
        idx1 = find_index(context, word1)
        replace_cand = gen_words(len(word1.split()))
        idx2 = find_index(context.replace(word1, replace_cand), word2)
    else:
        idx2 = find_index(context, word2)
        replace_cand = gen_words(len(word2.split()))
        idx1 = find_index(context.replace(word2, replace_cand), word1)
    
    return idx1, idx2


def mask(sentence: str, word: str) -> str:
    replaced = re.sub(rf'(?<![\w\/-])({word})(?=[^\w\/-])', '[MASK]', sentence)
    masked = ['[CLS]'] + [replaced] + ['[SEP]']
    return ' '.join(masked)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['minicons', 'minicons.bin']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.3.5,<2.0.0',
 'polyleven>=0.7,<0.8',
 'torch>=1.8.0,<2.0.0',
 'transformers>=4.4.1,<5.0.0',
 'urllib3>=1.26.7,<2.0.0']

entry_points = \
{'console_scripts': ['minicons = minicons.bin.score_cli:process']}

setup_kwargs = {
    'name': 'minicons',
    'version': '0.2.2',
    'description': 'A package of useful functions to analyze transformer based language models.',
    'long_description': '# minicons\n\n[![Downloads](https://static.pepy.tech/personalized-badge/minicons?period=total&units=international_system&left_color=black&right_color=brightgreen&left_text=Downloads)](https://pepy.tech/project/minicons)\n\nHelper functions for analyzing Transformer based representations of language\n\nThis repo is a wrapper around the `transformers` [library](https://huggingface.co/transformers) from hugging face :hugs:\n\n\n## Installation\n\nInstall from Pypi using:\n\n```pip install minicons```\n\n## Supported Functionality\n\n- Extract word representations from Contextualized Word Embeddings\n- Score sequences using language model scoring techniques, including masked language models following [Salazar et al. (2020)](https://www.aclweb.org/anthology/2020.acl-main.240.pdf).\n\n\n## Examples\n\n1. Extract word representations from contextualized word embeddings:\n\n```py\nfrom minicons import cwe\n\nmodel = cwe.CWE(\'bert-base-uncased\')\n\ncontext_words = [("I went to the bank to withdraw money.", "bank"), \n                 ("i was at the bank of the river ganga!", "bank")]\n\nprint(model.extract_representation(context_words, layer = 12))\n\n\'\'\' \ntensor([[ 0.5399, -0.2461, -0.0968,  ..., -0.4670, -0.5312, -0.0549],\n        [-0.8258, -0.4308,  0.2744,  ..., -0.5987, -0.6984,  0.2087]],\n       grad_fn=<MeanBackward1>)\n\'\'\'\n```\n\n2. Compute sentence acceptability measures (surprisals) using Word Prediction Models:\n\n```py\nfrom minicons import scorer\n\nmlm_model = scorer.MaskedLMScorer(\'bert-base-uncased\', \'cpu\')\nilm_model = scorer.IncrementalLMScorer(\'distilgpt2\', \'cpu\')\n\nstimuli = ["The keys to the cabinet are on the table.",\n           "The keys to the cabinet is on the table."]\n\n# use sequence_score with different reduction options: \n# Sequence Surprisal - lambda x: -x.sum(1)\n# Sequence Log-probability - lambda x: x.sum(1)\n# Sequence Surprisal, normalized by number of tokens - lambda x: -x.mean(1)\n# Sequence Log-probability, normalized by number of tokens - lambda x: x.mean(1)\n# and so on...\n\nprint(ilm_model.sequence_score(stimuli, reduction = lambda x: -x.sum(0).item()))\n\n\'\'\'\n[39.879737854003906, 42.75846481323242]\n\'\'\'\n\n# MLM scoring, inspired by Salazar et al., 2020\nprint(mlm_model.sequence_score(stimuli, reduction = lambda x: -x.sum(0).item()))\n\'\'\'\n[13.962685585021973, 23.415111541748047]\n\'\'\'\n```\n\n## Tutorials\n\n- [Introduction to using LM-scoring methods using minicons](https://kanishka.xyz/post/minicons-running-large-scale-behavioral-analyses-on-transformer-lms/)\n- [Computing sentence and token surprisals using minicons](examples/surprisals.md)\n- [Extracting word/phrase representations using minicons](examples/word_representations.md)\n\n## Recent Updates\n- **November 6, 2021:** MLM scoring has been fixed! You can now use `model.token_score()` and `model.sequence_score()` with `MaskedLMScorers` as well!',
    'author': 'Kanishka Misra',
    'author_email': 'kmisra@purdue.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kanishkamisra/minicons',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4',
}


setup(**setup_kwargs)



from logging import log
from typing import Iterable, Union, List, Dict, Optional, Callable, Tuple, Any

import torch
from transformers import AutoModelForCausalLM, AutoModelForMaskedLM, AutoTokenizer
from transformers.utils.logging import set_verbosity_error

from collections import defaultdict

from itertools import chain
from re import sub

import warnings

set_verbosity_error()

class LMScorer:
    """
    Base LM scorer class intended to store models and tokenizers along
    with methods to facilitate the analysis of language model output scores.
    """
    def __init__(self, model_name: str, device: Optional[str] = 'cpu') -> None:
        """
        :param model_name: name of the model, should either be a path
            to a model (.pt or .bin file) stored locally, or a
            pretrained model stored on the Huggingface Model Hub.
        :type model_name: str
        :param device: device type that the model should be loaded on,
            options: `cpu or cuda:{0, 1, ...}`
        :type device: str, optional
        """
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast = True)
        self.device = device
        self.vocab = defaultdict(list)
        # {self.vocab[x.strip()].append(i) for x, i in [(self.tokenizer.decode([i]), i) for i in range(self.tokenizer.vocab_size)]}
        for i in range(self.tokenizer.vocab_size):
            decoded = [(self.tokenizer.decode(i), i)]
            for x, j in decoded:
                self.vocab[x.strip()].append(j)

    def add_special_tokens(self, text: Union[str, List[str]]) -> Union[str, List[str]]:
        raise NotImplementedError
    
    def distribution(self, batch: Iterable) -> torch.Tensor:
        raise NotImplementedError
    
    def topk(self, distribution: torch.Tensor, k: int = 1) -> Tuple:
        top_k = distribution.topk(k)
    
        probs = top_k.values.squeeze(1).exp().tolist()
        if k == 1:
            tokens = self.decode(top_k.indices.squeeze(1))
        else:
            tokens = [self.decode(x) for x in top_k.indices.squeeze(1)]
    
        return tokens, probs

    def query(self, distribution: torch.Tensor, queries: List[str]) -> Tuple:
        # this will be self.vocab tho
        query_ids = [self.vocab[a] for a in queries]
        maxlen = max(map(len, query_ids))
        query_ids = [q + [self.tokenizer.pad_token_id] * (maxlen - len(q)) if len(q) < maxlen else q for q in query_ids]
        current_batch_size = distribution.shape[0]
        probs = distribution[torch.arange(current_batch_size)[:, None], query_ids].max(1).values.exp().tolist()
        
        inv_ranks = distribution.argsort().argsort() + 1
        ranks = distribution.shape[1] - inv_ranks + 1
        token_ranks = ranks[torch.arange(current_batch_size)[:, None], query_ids].min(1).values.tolist()
    
        return probs, token_ranks

    def logprobs(self, batch: Iterable, rank: bool = False) -> Union[float, List[float]]:
        warnings.warn(
            "logprobs is deprecated, use compute_stats instead",
            DeprecationWarning
        )
        raise NotImplementedError

    def compute_stats(self, batch: Iterable, rank: bool = False) -> Union[Union[float, int], List[Union[float, int]]]:
        raise NotImplementedError

    def prepare_text(self, text: Union[str, List[str]]) -> Union[str, List[str]]:
        raise NotImplementedError

    def prime_text(self, preamble: Union[str, List[str]], stimuli: Union[str, List[str]]) -> Tuple:
        raise NotImplementedError

    def token_score(self, batch: Union[str, List[str]], surprisal: bool = False, prob: bool = False, base_two: bool = False, rank: bool = False) -> Union[List[Tuple[str, float]], List[Tuple[str, float, int]]]:
        '''
        For every input sentence, returns a list of tuples in the following format:
            `(token, score)`,

        where score represents the log-probability (by default) of the token given context. Can also return ranks along with scores.

        :param ``Union[str, List[str]]`` batch: a single sentence or a batch of sentences.
        :param ``bool`` surprisal: If `True`, returns per-word surprisals instead of log-probabilities.
        :param ``bool`` prob: If `True`, returns per-word probabilities instead of log-probabilities.
        :param ``bool`` base_two: If `True`, uses log base 2 instead of natural-log (returns bits of values in case of surprisals)
        :param ``bool`` rank: If `True`, also returns the rank of each word in context (based on the log-probability value)

        :return: A `List` containing a `Tuple` consisting of the word, its associated score, and optionally, its rank.
        :rtype: ``Union[List[Tuple[str, float]], List[Tuple[str, float, int]]]``
        '''
        raise NotImplementedError
    
    def score(self, batch: Union[str, List[str]], pool: Callable = torch.mean, *args) -> Union[float, List[float]]:
        '''
        DEPRECATED as of v 0.1.18. Check out ``sequence_score`` or ``token_score`` instead!

        Pooled estimates of sentence log probabilities, computed by the
        language model. Pooling is usually done using a function that
        is passed to the method.

        :param batch: a list of sentences that will be passed to the
            language model to score.
        :type batch: Union[str, List[str]]
        :param pool: Pooling function, is selected to be
            `torch.mean()` by default.
        :type pool: Callable

        :return: Float or list of floats specifying the log
            probabilities of the input sentence(s).
        :rtype: Union[float, List[float]]
        '''
        warnings.warn(
            "score is deprecated, use sequence_score or token_score instead",
            DeprecationWarning
        )
        result = self.logprobs(self.prepare_text(batch))
        logprob, _ = list(zip(*result))
        pooled = list(map(lambda x: pool(x, *args).tolist(), logprob))
        
        return pooled
    
    def adapt_score(self, preamble: Union[str, List[str]], stimuli: Union[str, List[str]], pool: Callable = torch.mean, *args) -> None:
        """
        DEPRECATED as of v 0.1.18. Check out ``partial_score`` instead!
        """
        warnings.warn(
            "adapt_score is deprecated, use partial_score or token_score instead",
            DeprecationWarning
        )

    def partial_score(self, preamble: Union[str, List[str]], stimuli: Union[str, List[str]], reduction: Callable = lambda x: x.mean(0).item(), **kwargs) -> List[float]:
        '''
        Pooled estimates of sequence log probabilities (or some modification of it), given a preamble. Pooling is usually done using a function that is passed to the method.

        :param preamble: a batch of preambles or primes passed to the
            language model. This is what the sequence is conditioned on, and the model ignores the word probabilities of this part of the input in estimating the overall score.
        :type preamble: ``Union[str, List[str]]``
        :param stimuli: a batch of sequences (same length as preamble)
            that form the main input consisting of the sequence whose
            score you want to calculate.
        :type stimuli: ``Union[str, List[str]]``
        :param reduction: Reduction function, is selected to be
            ``lambda x: x.mean(0).item()`` by default, which stands for the avg. log-probability per token for each sequence in the batch.
        :type reduction: Callable
        :param kwargs: parameters for the ``compute_stats`` call --

            * `prob` (`bool`): Whether the returned value should be a probability (note that the default reduction method will have to be changed to `lambda x: x.prod(0).item()` to get a meaningful return value)

            * `base_two` (`bool`): whether the returned value should be in base 2 (only works when `prob = False`)

            * `surprisal` (`bool`): whether the returned value should be a surprisal (does not work when `prob = True`)


        :return: List of floats specifying the desired score for the stimuli part of the input, e.g., P(stimuli | preamble).
        :rtype: ``List[float]``
        '''
        result = self.compute_stats(self.prime_text(preamble, stimuli), **kwargs, return_tensors = True)
        logprob = result
        reduced = list(map(reduction, logprob))
        
        return reduced

    def encode(self, text: Union[str, List[str]], manual_special: bool = True, return_tensors: Optional[str] = 'pt') -> Dict:
        """
        Encode a batch of sentences using the model's tokenizer.
        Equivalent of calling `model.tokenizer(input)`

        :param ``Union[str, List[str]]`` text: Input batch/sentence to
            be encoded.
        :param manual_special: Specification of whether special tokens
            will be manually encoded.
        :type manual_special: bool
        :param return_tensors: returned tensor format. Default `'pt'`
        :type manual_special: str

        :return: Encoded batch 
        :rtype: ``Dict``
        """
        sentences = [text] if isinstance(text, str) else text

        if manual_special:
            # manually add special tokens
            sentences = self.add_special_tokens(sentences)
            if return_tensors:
                tokens = self.tokenizer.batch_encode_plus(sentences, add_special_tokens = False, padding = 'longest', return_attention_mask = True, return_tensors = return_tensors)
        else:
            # mostly for masked LMs
            tokens = self.tokenizer.batch_encode_plus(sentences, padding = 'longest', return_attention_mask = True)

        return tokens
    
    def decode(self, idx: List[int]):
        """
        Decode input ids using the model's tokenizer.

        :param ``List[int]`` idx: List of ids.

        :return: Decoded strings
        :rtype: List[str]
        """
        return [self.tokenizer.decode([x]).strip() for x in self.tokenizer.convert_tokens_to_ids(self.tokenizer.convert_ids_to_tokens(idx))]

class MaskedLMScorer(LMScorer):
    """
    Class for Masked Langauge Models such as BERT, RoBERTa, etc.

    :param model_name: name of the model, should either be a path
        to a model (.pt or .bin file) stored locally, or a
        pretrained model stored on the Huggingface Model Hub.
    :type model_name: str
    :param device: device type that the model should be loaded on,
        options: `cpu or cuda:{0, 1, ...}`
    :type device: str, optional
    """
    def __init__(self, model_name: str, device: Optional[str] = 'cpu') -> None:
        """
        :param model_name: name of the model, should either be a path
            to a model (.pt or .bin file) stored locally, or a
            pretrained model stored on the Huggingface Model Hub.

        :type model_name: str
        :param device: device type that the model should be loaded on,
            options: `cpu or cuda:{0, 1, ...}`
        :type device: str, optional
        """
        super(MaskedLMScorer, self).__init__(model_name, device)
        
        self.model = AutoModelForMaskedLM.from_pretrained(model_name, return_dict = True)
        self.model.to(self.device)
        self.model.eval()
        
        # define CLS and SEP tokens
        self.bos_token_id = self.tokenizer.cls_token_id
        self.eos_token_id = self.tokenizer.sep_token_id
        self.cls_token_id = self.tokenizer.cls_token_id
        self.sep_token_id = self.tokenizer.sep_token_id
        self.mask_token_id = self.tokenizer.mask_token_id
        self.pad_token_id = self.tokenizer.pad_token_id
    
    def add_special_tokens(self, text: Union[str, List[str]]) -> List[str]:
        """
        Reformats input text to add special model-dependent tokens.

        :param text: single string or batch of strings to be
            modified.
        :type text: ``Union[str, List[str]]``

        :return: Modified input, containing special tokens as per 
            tokenizer specification
        :rtype: ``List[str]``
        """
        sentences = [text] if isinstance(text, str) else text
        sentences = [self.tokenizer.cls_token + " " + sentence + " " + self.tokenizer.sep_token for sentence in sentences]

        return sentences

    def mask(self, sentence_words: Union[Tuple[str, str], List[Tuple[str, str]]]) -> Tuple[str, str, int]:
        """
        Processes a list of (sentence, word) into input that has the
        word masked out of the sentence. 
        
        Note: only works for masked LMs.

        :param ``Union[Tuple[str], List[Tuple[str]]]`` sentence_words:
            Input consisting of `[(sentence, word)]`, where sentence
            is an input sentence, and word is a word present in the
            sentence that will be masked out.

        :return: Tuple `(sentence, word, length)`
        """
        sentence_words = [sentence_words] if isinstance(sentence_words[0], str) else sentence_words
        sentences, words = list(zip(*sentence_words))
        words = list(words)
        length = len(words)

        sentences = [sub(rf'(?<![\w\/-])({word})(?=[^\w\/-])', self.tokenizer.mask_token, sentence) for sentence, word in sentence_words]

        return (sentences, words, length)

    def cloze(self, sentence_words: Union[Tuple[str, str], List[Tuple[str, str]]]) -> torch.Tensor:
        """
        Runs inference on masked input. 
        Note: only works for masked LMs.

        :param ``Union[Tuple[str], List[Tuple[str]]]`` sentence_words:
            Input consisting of `[(sentence, word)]`, where sentence
            is an input sentence, and word is a word present in the
            sentence that will be masked out and inferred.
        
        :return: A tensor with log probabilities for the desired word
            in context
        """
        sentences, words, length = self.mask(sentence_words)

        encoded = self.tokenizer(sentences, return_tensors='pt')
        encoded = encoded.to(self.device)

        idx = torch.nonzero(encoded['input_ids'] == self.tokenizer.mask_token_id, as_tuple=False)[:,1].unsqueeze(1)
        word_idx = self.tokenizer(words, add_special_tokens=False)['input_ids']
        with torch.no_grad():
            masked_logits = self.model(**encoded).logits[torch.arange(length)[:, None], idx].squeeze().detach()
            if len(sentences) > 1:
                logprobs = masked_logits - masked_logits.logsumexp(1).unsqueeze(1)
                masked_logprobs = logprobs[torch.arange(len(sentences))[:, None], word_idx].exp().squeeze()
            else:
                logprobs = masked_logits - masked_logits.logsumexp(0)
                masked_logprobs = logprobs[word_idx].exp().squeeze()

        return masked_logprobs


    def prepare_text(self, text: Union[str, List[str]]) -> Iterable[Any]:
        """
        Prepares a batch of input text into a format fit to run MLM
        scoring on. 

        Borrows preprocessing algorithm from Salazar et al. (2020), and
        modifies code from the following github repository by simonpri:
        https://github.com/simonepri/lm-scorer
        
        :param text: batch of sentences to be prepared for scoring.

        :return: Batch of formatted input that can be passed to `logprob`
        """
        # converts input text to batch of tensors with every position except the cls and sep token masked
        sentences = [text] if isinstance(text, str) else text
        
        # idea is to tokenize and then create batches of tokenized instances,
        # but with each token in the sequence replaced by the mask token. 
        
        encoded = self.encode(sentences, manual_special = False)

        token_idx = encoded['input_ids']
        attention_masks = encoded['attention_mask']

        masked_tensors = [] # token ids, attention masks, lengths

        for token_ids, attention_mask in zip(token_idx, attention_masks):
            token_ids = torch.tensor(token_ids)
            # final_lengths = len(token_ids) - 2
            attention_mask = torch.tensor(attention_mask)
            
            token_ids_masked_list = []
            attention_masked_list = []

            effective_token_ids = [token for token in token_ids if token != self.pad_token_id and token != self.cls_token_id and token != self.sep_token_id]
            effective_length = len(effective_token_ids)
            

            mask_indices = []
            mask_indices = [[mask_pos] for mask_pos in range(effective_length+2)]

            # We don't mask the [CLS], [SEP] for now for PLL
            mask_indices = mask_indices[1:-1]

            mask_token_id = self.mask_token_id
            for mask_set in mask_indices:
                token_ids_masked = token_ids.clone()
                token_ids_masked[mask_set] = mask_token_id
                attention_masked = attention_mask.clone()
                
                attention_masked_list.append(attention_masked)
                token_ids_masked_list.append(token_ids_masked)
            masked_tensors.append((torch.stack(token_ids_masked_list), torch.stack(attention_masked_list), effective_token_ids, len(mask_indices), 1))
        
        return masked_tensors

    def prime_text(self, preamble: Union[str, List[str]] , stimuli: Union[str, List[str]]) -> Iterable[Any]:
        """
        Prepares a batch of input text into a format fit to run LM
        scoring on. 

        Borrows preprocessing algorithm from Salazar et al. (2020), and
        modifies code from the following github repository by simonpri:
        https://github.com/simonepri/lm-scorer

        :param ``Union[str, List[str]]`` preamble: Batch of prefixes/prime/preambles on which the LM is conditioned.
        :param ``Union[str, List[str]]`` stimuli: Batch of continuations that are scored based on the conditioned text (provided in the ``preamble``). The positions of the elements match their counterparts in the ``preamble``.

        :return: Batch of formatted input that can be passed to
            ``compute_stats``
        """
        preamble_text = [preamble] if isinstance(preamble, str) else preamble
        preamble_encoded = self.encode(preamble_text, False)['input_ids']
        preamble_lens = []
        for preamble_tokens in preamble_encoded:
            preamble_lens.append(len([token for token in preamble_tokens if token != self.pad_token_id and token != self.sep_token_id]))
        
        sentences = [preamble + " " + stimuli] if isinstance(preamble, str) else [p + " " + s for p, s in list(zip(preamble, stimuli))]
            
        # idea is to tokenize and then create batches of tokenized instances,
        # but with each token in the sequence replaced by the mask token. 

        encoded = self.encode(sentences, manual_special = False)

        token_idx = encoded['input_ids']
        attention_masks = encoded['attention_mask']

        masked_tensors = [] # token ids, attention masks, lengths

        for i, (token_ids, attention_mask) in enumerate(zip(token_idx, attention_masks)):
            token_ids = torch.tensor(token_ids)
            # final_lengths = len(token_ids) - 2
            attention_mask = torch.tensor(attention_mask)

            token_ids_masked_list = []
            attention_masked_list = []
            
            effective_token_ids = [token for j, token in enumerate(token_ids) if token != self.pad_token_id and token != self.cls_token_id and token != self.sep_token_id and j >= preamble_lens[i]]
            effective_length = len(effective_token_ids) + preamble_lens[i]


            mask_indices = []
            mask_indices = [[mask_pos] for mask_pos in range(preamble_lens[i], effective_length+1)]

            # We don't mask the [CLS], [SEP] for now for PLL
            mask_indices = mask_indices[:-1]

            mask_token_id = self.mask_token_id
            for mask_set in mask_indices:
                token_ids_masked = token_ids.clone()
                token_ids_masked[mask_set] = mask_token_id
                attention_masked = attention_mask.clone()

                attention_masked_list.append(attention_masked)
                token_ids_masked_list.append(token_ids_masked)
            masked_tensors.append((torch.stack(token_ids_masked_list), torch.stack(attention_masked_list), effective_token_ids, len(mask_indices), preamble_lens[i]))

        return masked_tensors

    def distribution(self, batch: Iterable) -> torch.Tensor:
        """
        Returns a distribution over the vocabulary of the model.

        :param `Iterable` batch: A batch of inputs fit to pass to a
            transformer LM.

        :return: Tensor consisting of log probabilies over vocab items.
        """
        # takes in prepared text and returns scores for each sentence in batch
        token_ids, attention_masks, effective_token_ids, lengths, offsets = list(zip(*batch))
        token_ids = torch.cat(token_ids)
        attention_masks = torch.cat(attention_masks)
        token_ids = token_ids.to(self.device)
        attention_masks = attention_masks.to(self.device)
        effective_token_ids = torch.cat([torch.tensor(x) for x in effective_token_ids])

        indices = list(chain.from_iterable([list(range(o,o+n)) for n, o in zip(lengths, offsets)]))
        with torch.no_grad():
            output = self.model(token_ids, attention_mask = attention_masks)
            logits = output.logits[torch.arange(sum(lengths)), indices].detach()

        logprob_distribution = logits - logits.logsumexp(1).unsqueeze(1)

        return logprob_distribution

    def cloze_distribution(self, queries: Iterable) -> torch.Tensor:
    
        '''
        Accepts as input batch of [(s_i, bw_i)] where s_i is a prompt with an
        abstract token (bw_i) representing a blank word and returns a distribution
        over the vocabulary of the model.

        :param `Iterable` queries: A batch of [(s_i, bw_i)] where s_i is a prompt with an abstract token (bw_i) representing a blank word

        :return: Tensor contisting of log probabilities over vocab items.
        '''
        
        queries = [queries] if isinstance(queries[0], str) else queries
        prompts, words = list(zip(*queries))
            
        modified_prompts = self.add_special_tokens(prompts)
        splits = [prompt.split(word) for prompt, word in zip(modified_prompts, words)]
        splits = [[x.strip() for x in s] for s in splits]
        pre, post = list(zip(*splits))
        pre_idx = self.tokenizer(list(pre), add_special_tokens = False, padding=False)['input_ids']
        mask_idx = [len(item) for item in pre_idx]
        masked = [m.replace(w, self.tokenizer.mask_token) for m, w in zip(modified_prompts, words)]
        
        with torch.no_grad():
            encoded = self.tokenizer(masked, add_special_tokens = False, return_tensors='pt', padding = True)
            encoded = encoded.to(self.device)
            logits = self.model(**encoded)
            presoftmax = logits.logits[torch.arange(len(queries)), mask_idx]
            if 'cuda' in self.device:
                presoftmax = presoftmax.detach().cpu()
            else:
                presoftmax = presoftmax.detach()
            
        logprobs = presoftmax - presoftmax.logsumexp(1).unsqueeze(1)
        
        return logprobs    

    def logprobs(self, batch: Iterable, rank = False) -> Union[List[Tuple[torch.Tensor, str]], List[Tuple[torch.Tensor, str, int]]]:
        """
        Returns log probabilities

        :param `Iterable` batch: A batch of inputs fit to pass to a
            transformer LM.
        :param rank: Specifies whether to also return ranks of words.
        :type rank: bool

        :return: List of MLM score metrics and tokens.
        :rtype: Union[List[Tuple[torch.Tensor, str]], List[Tuple[torch.Tensor, str, int]]]
        """
        warnings.warn(
            "logprobs is deprecated, use compute_stats instead",
            DeprecationWarning
        )
        token_ids, attention_masks, effective_token_ids, lengths, offsets = list(zip(*batch))
        token_ids = torch.cat(token_ids)
        attention_masks = torch.cat(attention_masks)
        token_ids = token_ids.to(self.device)
        attention_masks = attention_masks.to(self.device)
        effective_token_ids = torch.cat([torch.tensor(x) for x in effective_token_ids])

        sent_tokens = list(map(lambda x: self.tokenizer.convert_ids_to_tokens(x.tolist()), effective_token_ids.split(lengths)))
        
        indices = list(chain.from_iterable([list(range(o,o+n)) for n, o in zip(lengths, offsets)]))
        with torch.no_grad():
            output = self.model(token_ids, attention_mask = attention_masks)
            logits = output.logits[torch.arange(sum(lengths)), indices]
            if self.device == 'cuda:0' or self.device == "cuda:1":
                logits.detach()
            
            sent_log_probs = logits - logits.logsumexp(1).unsqueeze(1)
            if rank:
                shape = sent_log_probs.shape
                # inv_ranks = (sent_log_probs).argsort().argsort() + 1
                # ranks = shape[1] - inv_ranks + 1
                ranks = (-1.0 * sent_log_probs).argsort().argsort() + 1
                word_ranks = ranks[torch.arange(shape[0]), effective_token_ids].split(lengths)
            sent_log_probs = sent_log_probs[torch.arange(sum(lengths)), effective_token_ids].type(torch.DoubleTensor).split(lengths)
            # print(sent_log_probs)
            # sentence_scores = list(map(lambda x: x.sum().tolist(), logprobs))
            # outputs.append((logprobs, sent_tokens))
            if rank:
                return list(zip(sent_log_probs, sent_tokens, word_ranks))
            
        return list(zip(sent_log_probs, sent_tokens))

    def compute_stats(self, batch: Iterable, rank: bool = False, prob = False, base_two: bool = False, return_tensors: bool = False) -> Union[Tuple[List[float], List[float]], List[float]]:
        '''
        Primary computational method that processes a batch of prepared sentences and returns per-token scores for each sentence. By default, returns log-probabilities.

        :param ``Iterable`` batch: batched input as processed by ``prepare_text`` or ``prime_text``.
        :param ``bool`` rank: whether the model should also return ranks per word (based on the conditional log-probability of the word in context).
        :param ``bool`` prob: whether the model should return probabilities instead of log-probabilities. Can only be `True` when `base_two` is `False`.
        :param ``bool`` base_two: whether the base of the log should be 2 (usually preferred when reporting results in bits). Can only be `True` when `prob` is `False`.
        :param ``bool`` return_tensors: whether the model should return scores as a list of tensors instead of a list of lists. This is important in some other convenient methods used in the package.

        :return: Either a tuple of lists, each containing probabilities and ranks per token in each sentence passed in the input.
        :rtype: ``Union[Tuple[List[float], List[float]], List[float]]``
        '''
        assert not (base_two and prob), "cannot both use base (which is for a log), and a probability measure at the same time!"

        token_ids, attention_masks, effective_token_ids, lengths, offsets = list(zip(*batch))
        token_ids = torch.cat(token_ids)
        attention_masks = torch.cat(attention_masks)
        token_ids = token_ids.to(self.device)
        attention_masks = attention_masks.to(self.device)
        effective_token_ids = torch.cat([torch.tensor(x) for x in effective_token_ids])
        
        indices = list(chain.from_iterable([list(range(o,o+n)) for n, o in zip(lengths, offsets)]))

        with torch.no_grad():
            output = self.model(token_ids, attention_mask = attention_masks)
            logits = output.logits.detach()[torch.arange(sum(lengths)), indices]

        logprob_distribution = logits - logits.logsumexp(1).unsqueeze(1)

        if base_two:
            logprob_distribution = logprob_distribution/torch.tensor(2).log()

        if prob:
            logprob_distribution = logprob_distribution.exp()

        if rank:
            shape = logprob_distribution.shape
            '''
            Double argsort trick:
            first argsort returns idxes of values that would return a sorted tensor,
            second argsort returns ranks (0 indexed)

            Proof: https://www.berkayantmen.com/rank.html

            TODO: Try to implement ranking in linear time but across arbitrary dimensions:
            https://stackoverflow.com/a/5284703
            '''
            word_ranks = (-1.0 * logprob_distribution).argsort().argsort() + 1
            word_ranks = word_ranks[torch.arange(shape[0]), effective_token_ids].split(lengths)
            word_ranks = [wr.tolist() for wr in word_ranks]

        scores = logprob_distribution[torch.arange(sum(lengths)), effective_token_ids].type(torch.DoubleTensor).split(lengths)
        scores = [s for s in scores]

        if not return_tensors:
            scores = [s.tolist() for s in scores]

        if rank:
            return scores, word_ranks
        else:
            return scores

    def sequence_score(self, batch, reduction = lambda x: x.mean(0).item(), base_two = False):
        '''
        TODO: reduction should be a string, if it's a function, specify what kind of function. --> how to ensure it is always that type?
        '''
        tokenized = self.prepare_text(batch)
        scores = self.compute_stats(tokenized, rank = False, base_two = base_two, return_tensors = True)
        reduced = list(map(reduction, scores))
        return reduced

    def token_score(self, batch: Union[str, List[str]], surprisal: bool = False, prob: bool = False, base_two: bool = False, rank: bool = False) -> Union[List[Tuple[str, float]], List[Tuple[str, float, int]]]:
        '''
        For every input sentence, returns a list of tuples in the following format:
            `(token, score)`,

        where score represents the log-probability (by default) of the token given context. Can also return ranks along with scores.

        :param ``Union[str, List[str]]`` batch: a single sentence or a batch of sentences.
        :param ``bool`` surprisal: If `True`, returns per-word surprisals instead of log-probabilities.
        :param ``bool`` prob: If `True`, returns per-word probabilities instead of log-probabilities.
        :param ``bool`` base_two: If `True`, uses log base 2 instead of natural-log (returns bits of values in case of surprisals)
        :param ``bool`` rank: If `True`, also returns the rank of each word in context (based on the log-probability value)

        :return: A `List` containing a `Tuple` consisting of the word, its associated score, and optionally, its rank.
        :rtype: ``Union[List[Tuple[str, float]], List[Tuple[str, float, int]]]``
        '''
        assert not (surprisal and prob), "cannot both evaluate probability and surprisal at the same time!"
        assert not (base_two and prob), "cannot both use base (which is for a log), and a probability measure at the same time!"

        tokenized = self.prepare_text(batch)
        if rank:
            scores, ranks = self.compute_stats(tokenized, rank = rank, prob = prob, base_two = base_two, return_tensors=True)
        else:
            scores = self.compute_stats(tokenized, prob = prob, base_two = base_two, return_tensors=True)

        if surprisal:
            scores = [-1.0 * s for s in scores]

        scores = [s.tolist() for s in scores]

        indices = [[i.item() for i in indexed if i.item() != self.tokenizer.pad_token_id] for indexed in list(zip(*tokenized))[2]]
        tokens = [self.decode(idx) for idx in indices]

        if rank:
            assert len(tokens) == len(scores) == len(ranks)
        else:
            assert len(tokens) == len(scores)

        res = []
        if rank:
            for t, s, r in zip(tokens, scores, ranks):
                res.append(list(zip(t, s, r)))
            # return [list(zip(t, s, r)) for t, s, r in zip(tokens, scores, ranks)]
        else:
            for t, s in zip(tokens, scores):
                res.append(list(zip(t, s)))

        return res

class IncrementalLMScorer(LMScorer):
    """
    Class for Autoregressive or Incremental (or left-to-right) language models such as GPT2, etc.

    :param model_name: name of the model, should either be a path
        to a model (.pt or .bin file) stored locally, or a
        pretrained model stored on the Huggingface Model Hub.
    :type model_name: str
    :param device: device type that the model should be loaded on,
        options: `cpu or cuda:{0, 1, ...}`
    :type device: str, optional
    """
    def __init__(self, model_name: str, device: Optional[str] = 'cpu') -> None:
        """
        :param model_name: name of the model, should either be a path
            to a model (.pt or .bin file) stored locally, or a
            pretrained model stored on the Huggingface Model Hub.

        :type model_name: str
        :param device: device type that the model should be loaded on,
            options: `cpu or cuda:{0, 1, ...}`
        :type device: str, optional
        """
        super(IncrementalLMScorer, self).__init__(model_name, device)
        
        self.model = AutoModelForCausalLM.from_pretrained(model_name, return_dict = True)
        
        # define CLS and SEP tokens
        if self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"additional_special_tokens": ["<|pad|>"]})
            self.tokenizer.pad_token = "<|pad|>"

        if self.tokenizer.bos_token is None:
            self.tokenizer.add_special_tokens({"additional_special_tokens": ["<|bos|>"]})
            self.tokenizer.bos_token = "<|bos|>"

        self.model.resize_token_embeddings(len(self.tokenizer))
        self.model.to(self.device)
        self.model.eval()
    
    def add_special_tokens(self, text: Union[str, List[str]]) -> Union[str, List[str]]:
        """
        Reformats input text to add special model-dependent tokens.

        :param text: single string or batch of strings to be
            modified.
        :type text: Union[str, List[str]]
        
        :return: Modified input, containing special tokens as per 
            tokenizer specification
        :rtype: Union[float, List[float]]:
        """
        sentences = [text] if isinstance(text, str) else text
        sentences = [self.tokenizer.bos_token + sentence for sentence in sentences]

        return sentences

    def encode(self, text: Union[str, List[str]]) -> dict:
        text = [text] if isinstance(text, str) else text
        return self.tokenizer(text, return_tensors='pt', padding = True)
    
    def prepare_text(self, text: Union[str, List[str]]) -> Tuple:
        """
        Prepares a batch of input text into a format fit to run LM
        scoring on. 

        :param text: batch of sentences to be prepared for scoring.
        
        :return: Batch of formatted input that can be passed to
            ``compute_stats``
        """
        encoded = self.encode(text)
        offsets = [0] * len(encoded['input_ids'])
        return encoded, offsets
    
    def prime_text(self, preamble: Union[str, List[str]], stimuli: Union[str, List[str]]) -> Tuple:
        """
        Prepares a batch of input text into a format fit to run LM
        scoring on. 

        :param ``Union[str, List[str]]`` preamble: Batch of prefixes/prime/preambles on which the LM is conditioned.
        :param ``Union[str, List[str]]`` stimuli: Batch of continuations that are scored based on the conditioned text (provided in the ``preamble``). The positions of the elements match their counterparts in the ``preamble``.

        :return: Batch of formatted input that can be passed to
            ``compute_stats``
        """
        preamble_text = [preamble] if isinstance(preamble, str) else preamble
        preamble_encoded = self.tokenizer(preamble_text)['input_ids']
        preamble_lens = []
        for preamble_tokens in preamble_encoded:
            preamble_lens.append(len([token for token in preamble_tokens if token != self.tokenizer.pad_token_id and token != self.tokenizer.sep_token_id]) - 1)
        
        sentences = [preamble + " " + stimuli] if isinstance(preamble, str) else [p + " " + s for p , s in list(zip(preamble, stimuli))]
            
        return self.encode(sentences), preamble_lens
    
    def distribution(self, batch: Iterable) -> torch.Tensor:
        """
        Returns a distribution over the vocabulary of the model.

        :param `Iterable` batch: A batch of inputs fit to pass to a
            transformer LM.

        :return: Tensor consisting of log probabilies over vocab items.
        """
        batch, offsets = batch
        ids = batch["input_ids"]
        ids = ids.to(self.device)
        attention_masks = batch["attention_mask"]
        attention_masks = attention_masks.to(self.device)
        nopad_mask = ids != self.tokenizer.pad_token_id

        with torch.no_grad():
            outputs = self.model(ids, attention_mask=attention_masks)
            logits = outputs.logits
            if self.device == 'cuda:0' or self.device == "cuda:1":
                logits.detach()

        outputs = []
        for sent_index in range(len(ids)):
            sent_nopad_mask = nopad_mask[sent_index]
            # len(tokens) = len(text[sent_index]) + 1
            sent_tokens = [
                tok
                for i, tok in enumerate(batch.tokens(sent_index))
                if sent_nopad_mask[i] and i > offsets[sent_index] + 1
            ]

            # sent_ids.shape = [len(text[sent_index]) + 1]
            # ignore first token (<|eos|>)
            sent_ids = ids[sent_index, sent_nopad_mask][1:]
            # logits.shape = [len(text[sent_index]) + 1, vocab_size]
            sent_logits = logits[sent_index, sent_nopad_mask][:-1, :]
            sent_logits[:, self.tokenizer.pad_token_id] = float("-inf")

            outputs.append(sent_logits[-1])
        return torch.stack(outputs, 0)

    def next_word_distribution(self, queries: List, surprisal: bool = False):
        '''
        Returns the log probability distribution of the next word.
        '''
        encoded = self.encode(queries)
        encoded = encoded.to(self.device)
        query_ids = [[j for j, i in enumerate(instance) if i != self.tokenizer.pad_token_id][-1] for instance in encoded['input_ids'].tolist()]

        logits = self.model(**encoded).logits.detach()
        logits[:, :, self.tokenizer.pad_token_id] = float("-inf")

        logits = logits[torch.arange(len(query_ids)), query_ids]
        logprobs = logits - logits.logsumexp(1).unsqueeze(1)

        if surprisal:
            logprobs = -1.0 * logprobs
        
        return logprobs

    def compute_stats(self, batch: Iterable, rank: bool = False, prob: bool = False, base_two: bool = False, return_tensors: bool = False) -> Union[Tuple[List[float], List[float]], List[float]]:
        '''
        Primary computational method that processes a batch of prepared sentences and returns per-token scores for each sentence. By default, returns log-probabilities.

        :param ``Iterable`` batch: batched input as processed by ``prepare_text`` or ``prime_text``.
        :param ``bool`` rank: whether the model should also return ranks per word (based on the conditional log-probability of the word in context).
        :param ``bool`` prob: whether the model should return probabilities instead of log-probabilities. Can only be `True` when `base_two` is `False`.
        :param ``bool`` base_two: whether the base of the log should be 2 (usually preferred when reporting results in bits). Can only be `True` when `prob` is `False`.
        :param ``bool`` return_tensors: whether the model should return scores as a list of tensors instead of a list of lists. This is important in some other convenient methods used in the package.

        :return: Either a tuple of lists, each containing probabilities and ranks per token in each sentence passed in the input.
        :rtype: ``Union[Tuple[List[float], List[int]], List[float]]``
        '''
        assert not (base_two and prob), "cannot both use base (which is for a log), and a probability measure at the same time!"

        encoded, offsets = batch
        encoded = encoded.to(self.device)
        
        ids = [[i for i in instance if i != self.tokenizer.pad_token_id] for instance in encoded['input_ids'].tolist()]

        ## Ignore the probabilities of the first token.
        effective_ids = [id[1:] for id in ids]

        with torch.no_grad():
            logits = self.model(**encoded).logits.detach()

        logits[:, :, self.tokenizer.pad_token_id] = float("-inf")

        logits = logits.split([1]*len(offsets))

        ## Set up storage variables
        scores = []
        if rank:
            ranks = []

        for logit, idx, offset in zip(logits, effective_ids, offsets):
            length = len(idx)
            logit = logit.squeeze(0)[:, :-1][torch.arange(offset, length),]

            logprob_distribution = logit - logit.logsumexp(1).unsqueeze(1)
            query_ids = idx[offset:]
            if base_two:
                '''
                Log_2(X) = log_e(X)/log_e(2) (broadcasted)
                '''
                score = (logprob_distribution[torch.arange(length - offset), query_ids] / torch.tensor(2).log()).tolist()
            else:
                if prob:
                    score = logprob_distribution[torch.arange(length - offset), query_ids].exp().tolist()
                else:
                    score = logprob_distribution[torch.arange(length - offset), query_ids].tolist()

            if rank:
                # shape = logprob_distribution.shape
                '''
                Double argsort trick:
                first argsort returns idxes of values that would return a sorted tensor,
                second argsort returns ranks (0 indexed)

                Proof: https://www.berkayantmen.com/rank.html

                TODO: Try to implement ranking in linear time but across arbitrary dimensions:
                https://stackoverflow.com/a/5284703
                '''
                word_ranks = (-1.0 * logprob_distribution).argsort().argsort() + 1
                # inv_ranks = logprob_distribution.argsort().argsort() + 1
                # word_ranks = shape[1] - inv_ranks + 1
                word_ranks = word_ranks[torch.arange(length - offset), query_ids].tolist()
                ranks.append(word_ranks)

            scores.append(score)

        if return_tensors:
            scores = [torch.tensor(l) for l in scores]

        if rank:
            return scores, ranks
        else:
            return scores

    def sequence_score(self, batch, reduction = lambda x: x.mean(0).item(), base_two = False):
        '''
        TODO: reduction should be a string, if it's a function, specify what kind of function. --> how to ensure it is always that type?
        '''
        tokenized = self.prepare_text(batch)
        scores = self.compute_stats(tokenized, rank = False, base_two = base_two, return_tensors = True)
        reduced = list(map(reduction, scores))
        return reduced

    def token_score(self, batch: Union[str, List[str]], surprisal: bool = False, prob: bool = False, base_two: bool = False, rank: bool = False) -> Union[List[Tuple[str, float]], List[Tuple[str, float, int]]]:
        '''
        For every input sentence, returns a list of tuples in the following format:
            `(token, score)`,

        where score represents the log-probability (by default) of the token given context. Can also return ranks along with scores.

        :param ``Union[str, List[str]]`` batch: a single sentence or a batch of sentences.
        :param ``bool`` surprisal: If `True`, returns per-word surprisals instead of log-probabilities.
        :param ``bool`` prob: If `True`, returns per-word probabilities instead of log-probabilities.
        :param ``bool`` base_two: If `True`, uses log base 2 instead of natural-log (returns bits of values in case of surprisals)
        :param ``bool`` rank: If `True`, also returns the rank of each word in context (based on the log-probability value)

        :return: A `List` containing a `Tuple` consisting of the word, its associated score, and optionally, its rank.
        :rtype: ``Union[List[Tuple[str, float]], List[Tuple[str, float, int]]]``
        '''

        assert not (surprisal and prob), "cannot both evaluate probability and surprisal at the same time!"
        assert not (base_two and prob), "cannot both use base (which is for a log), and a probability measure at the same time!"

        tokenized = self.prepare_text(batch)
        if rank:
            scores, ranks = self.compute_stats(tokenized, rank = rank, prob = prob, base_two = base_two, return_tensors=True)
        else:
            scores = self.compute_stats(tokenized, prob = prob, base_two = base_two, return_tensors=True)

        if surprisal:
            scores = [-1.0 * s for s in scores]

        scores = [s.tolist() for s in scores]

        indices = [[i for i in indexed if i != self.tokenizer.pad_token_id] for indexed in tokenized[0]['input_ids'].tolist()]
        tokens = [self.decode(idx) for idx in indices]

        if rank:
            assert len(tokens) == len(scores) == len(ranks)
        else:
            assert len(tokens) == len(scores)

        res = []
        if rank:
            for t, s, r in zip(tokens, scores, ranks):
                if len(t) > len(s):
                    diff = len(t) - len(s)
                    sc = [0.0]*diff + s
                    ra = [0]*diff + r
                    res.append(list(zip(t, sc, ra)))
                else:
                    res.append(list(zip(t, sc, ra)))
            # return [list(zip(t, s, r)) for t, s, r in zip(tokens, scores, ranks)]
        else:
            for t, s in zip(tokens, scores):
                if len(t) > len(s):
                    diff = len(t) - len(s)
                    sc = [0.0]*diff + s
                    res.append(list(zip(t, sc)))
                else:
                    res.append(list(zip(t, sc)))

        return res

    def logprobs(self, batch: Iterable, rank = False) -> Union[float, List[float]]:
        """
        Returns log probabilities

        :param `Iterable` batch: A batch of inputs fit to pass to a
            transformer LM.
        :param rank: Specifies whether to also return ranks of words.
        :type rank: bool

        :return: List of LM score metrics (probability and rank)
            and tokens.
        :rtype: Union[List[Tuple[torch.Tensor, str]], List[Tuple[torch.Tensor, str, int]]]
        """
        warnings.warn(
            "logprobs is deprecated, use compute_stats instead",
            DeprecationWarning
        )
        batch, offsets = batch
        ids = batch["input_ids"]
        ids = ids.to(self.device)
        attention_masks = batch["attention_mask"]
        attention_masks = attention_masks.to(self.device)
        nopad_mask = ids != self.tokenizer.pad_token_id

        with torch.no_grad():
            outputs = self.model(ids, attention_mask=attention_masks)
            logits = outputs.logits
            if self.device == 'cuda:0' or self.device == "cuda:1":
                logits.detach()
        
        outputs = []
        for sent_index in range(len(ids)):
            sent_nopad_mask = nopad_mask[sent_index]
            # len(tokens) = len(text[sent_index]) + 1
            sent_tokens = [
                tok
                for i, tok in enumerate(batch.tokens(sent_index))
                if sent_nopad_mask[i] and i > offsets[sent_index]
            ]

            # sent_ids.shape = [len(text[sent_index]) + 1]
            # ignore first token (<|eos|>)
            sent_ids = ids[sent_index, sent_nopad_mask][1:]
            # logits.shape = [len(text[sent_index]) + 1, vocab_size]
            sent_logits = logits[sent_index, sent_nopad_mask][:-1, :]
            sent_logits[:, self.tokenizer.pad_token_id] = float("-inf")
            # ids_scores.shape = [seq_len + 1]
            # select only the ids present in the sentence out of all vocab items (as a 2d array)
            sent_ids_scores = sent_logits.gather(1, sent_ids.unsqueeze(1)).squeeze(1)
            # log_prob.shape = [seq_len + 1]
            sent_log_probs = sent_ids_scores - sent_logits.logsumexp(1)
            
            sent_log_probs = sent_log_probs.type(torch.DoubleTensor)
            sent_log_probs = sent_log_probs[offsets[sent_index]:]
            lengths = len(sent_log_probs)
            if rank:
                shape = sent_logits.shape
                inv_ranks = (sent_logits).argsort().argsort() + 1
                ranks = shape[1] - inv_ranks + 1
                word_ranks = ranks[list(range(shape[0]))[offsets[sent_index]:], sent_ids[offsets[sent_index]: ].tolist()].split(lengths)
                word_ranks = [x[0] for x in word_ranks]
                outputs.append((sent_log_probs, sent_tokens, word_ranks))
            else:
                outputs.append((sent_log_probs, sent_tokens))
            # output = (sent_log_probs.sum(), sent_ids, sent_tokens)
            # outputs.append(output)
        return outputs